# TODO eh. confusing how guess and auto are different...
from typing import Iterable, Any

from ..common import Extraction, PathIsh


def is_git_repo(p: str) -> bool:
    if '://github.com/' in p:
        return True
    return False


def is_website(p: str) -> bool:
    if p.startswith('http'):
        return True
    return False


def index(path: PathIsh, *args, **kwargs) -> Iterable[Extraction]:
    ps = str(path)
    # TODO better url detection

    index_: Any # meh
    if is_git_repo(ps):
        from . import vcs
        index_ = vcs.index
    elif is_website(ps):
        from . import website
        index_ = website.index
    else:
        from . import auto
        index_ = auto.index
    yield from index_(path, *args, **kwargs)



from pathlib import Path
from typing import Dict, List, Tuple
import shutil

from sqlalchemy import create_engine, MetaData # type: ignore
from sqlalchemy import Column, Table # type: ignore

from cachew import NTBinder, ichunks

from .common import get_logger, DbVisit, get_tmpdir
from . import config


def dump_histories(all_histories: List[Tuple[str, List[DbVisit]]]):
    logger = get_logger()
    output_dir = Path(config.get().OUTPUT_DIR)
    db_path = output_dir / 'promnesia.sqlite'

    def iter_visits():
        for e, h in all_histories:
            # TODO sort them somehow for determinism?
            # TODO what do we do with errors?
            # TODO maybe conform them to schema and dump too?
            # TODO or, dump to a separate table?
            yield from h

    tpath = Path(get_tmpdir().name) / 'promnesia.tmp.sqlite'
    engine = create_engine(f'sqlite:///{tpath}')
    binder = NTBinder.make(DbVisit)
    meta = MetaData(engine)
    table = Table('visits', meta, *binder.columns)
    meta.create_all()

    with engine.begin() as conn:
        for chunk in ichunks(iter_visits(), n=1000):
            bound = [binder.to_row(x) for x in chunk]
            # pylint: disable=no-value-for-parameter
            conn.execute(table.insert().values(bound))

    shutil.move(str(tpath), str(db_path))

    logger.info('saved database to %s', db_path)



from pathlib import Path
import re
from subprocess import check_call
from typing import Iterable

from ..common import Extraction, PathIsh, get_tmpdir, slugify


def index(path: PathIsh, *args, **kwargs) -> Iterable[Extraction]:
    repo = str(path)

    # TODO this looks pretty horrible as a context name
    # perhaps pass context here since we know it should be github repo?
    tp = Path(get_tmpdir().name) / slugify(repo)
    check_call(['git', 'clone', repo, tp])

    def replacer(p: PathIsh, prefix=str(tp), repo=repo) -> str:
        ps = str(p)
        # TODO prefix is a bit misleading
        pos = ps.find(prefix)
        if pos == -1:
            # TODO not sure if should happen...
            return ps
        # TODO ugh. seems that blame view https://github.com/davidgasquez/handbook/blame/master/README.md#L25 is the most reliable
        # in raw mode can't jump onto line, when markdown is renderend can't jump either
        rest = ps[pos + len(prefix):]
        rest = re.sub(r':(\d+)$', r'#L\1', rest) # patch line number...
        return repo + '/blame/master' + rest

        # TODO doesn't work for git:
        # TODO think about something more generic... this isn't too sustainable
    # TODO not sure if context should be local or github?...

    from . import auto
    yield from auto.index(tp, *args, replacer=replacer, **kwargs)



"""
Tries to index as much as it can:

- TODO walks up the filesystem hierarchy
- guesses org/markdown/json/etc by extension or mime type

"""

import csv
from datetime import datetime
import json
from typing import Optional, Iterable, Union, List, Tuple, NamedTuple, Sequence, Iterator, Callable, Any, Dict
from fnmatch import fnmatch
from pathlib import Path
from functools import lru_cache, wraps

import pytz

from ..common import Visit, Url, PathIsh, get_logger, Loc, get_tmpdir, extract_urls, Extraction


@lru_cache(1)
def _magic():
    import magic # type: ignore
    return magic.Magic(mime=True)


def mime(path: PathIsh) -> str:
    return _magic().from_file(str(path))

Ctx = Sequence[str]

class EUrl(NamedTuple):
    url: Url
    ctx: Ctx # TODO ctx here is more like a Loc


def _collect(thing, path: List[str], result: List[EUrl]):
    if isinstance(thing, str):
        ctx: Ctx = tuple(path) # type: ignore
        result.extend([EUrl(url=u, ctx=ctx) for u in extract_urls(thing)])
    elif isinstance(thing, list):
        path.append('[]')
        for x in thing:
            _collect(x, path, result)
        path.pop()
    elif isinstance(thing, dict):
        for k, v in thing.items():
            path.append(k)
            _collect(k, path, result)
            _collect(v, path, result)
            path.pop()
    else:
        pass


# TOOD mm. okay, I suppose could use kython consuming thingy?..
def collect_from(thing) -> List[EUrl]:
    uuu: List[EUrl] = []
    path: List[str] = []
    _collect(thing, path, uuu)
    return uuu

# TODO use magic and mimes maybe?


Urls = Iterator[EUrl]

def _csv(path: Path) -> Urls:
    # TODO these could also have Loc to be fair..
    with path.open() as fo:
        # TODO shit need to urldecode
        reader = csv.DictReader(fo)
        for line in reader:
            yield from collect_from(line)


def _json(path: Path) -> Urls:
    jj = json.loads(path.read_text())
    yield from collect_from(jj)


def _plaintext(path: Path) -> Iterator[Extraction]:
    from . import shellcmd
    from .plaintext import extract_from_path
    logger = get_logger()
    # TODO eh? shellcmd?
    yield from shellcmd.index(extract_from_path(path))


# TODO think about the type
# TODO could pass fallback reason to the results as well?
def fallback(ex):
    """Falls back to plaintext in case of issues"""
    @wraps(ex)
    def wrapped(path: Path):
        try:
            it = ex(path)
            # ugh. keeping yeild in the try section is not ideal, but need this because of lazy yield semantics
            yield from it
        except ModuleNotFoundError as me:
            logger = get_logger()
            logger.exception(me)
            # TODO maybe check first? That way could sync it with the dependencies too
            logger.warning('%s not found, so falling back to plaintext! "pip3 install --user %s" for better support!', me.name, me.name)
            yield me
            yield from _plaintext(path)
    return wrapped


@fallback
def _markdown(path: Path) -> Iterator[Extraction]:
    from . import markdown
    yield from markdown.extract_from_file(path)


@fallback
def _html(path: Path) -> Iterator[Extraction]:
    from . import html
    yield from html.extract_from_file(path)


@fallback
def _org(path: Path) -> Iterator[Extraction]:
    from . import org
    return org.extract_from_file(path)


SMAP = {
    'application/json': _json,
    '.json'           : _json,

    '.csv'        : _csv,

    '.org'        : _org,
    '.org_archive': _org,

    '.md'         : _markdown,
    '.markdown'   : _markdown,

    'text/plain'  : _plaintext,
    '.txt'        : _plaintext,
    '.page'       : _plaintext,
    '.rst'        : _plaintext,


    # TODO doesn't work that great; weird stuff like
    # builtins.ImportError.name|2019-07-10T12:12:35.584510+00:00|names::ImportError::node::names::name::node::fullname
    # TODO could have stricter url extraction for that; always using http/https?
    # '.ipynb'      : _json,

    '.html'    : _html,
    'text/html': _html,


    # TODO not sure about these:
    'text/xml': None,
    'text/x-python': None,
    'text/x-tex': None,
    'text/x-lisp': None,
    'text/x-shellscript': None,
    'text/x-java': None,
    'text/troff': None,
    'text/x-c': None,
    'text/x-c++': None,
    'text/x-makefile': None,
    # TODO could reuse magic lib?

    # TODO def could extract from source code...
    '.tex': None, # TODO not sure..
    '.css': None,
    '.sh' : None,
    '.js' : None,
    '.hs' : None,
    '.bat': None,
    '.pl' : None,
    '.h'  : None,
    '.rs' : None,


    # TODO possible in theory?
    '.ppt' : None,
    '.pptx': None,
    '.xlsx': None,
    '.doc' : None,
    '.docx': None,
    '.ods' : None,
    '.odt' : None,
    '.rtf' : None,
    '.epub': None,
    '.pdf' : None,
    '.vcf' : None,
    '.djvu': None,
    '.dvi' : None,
    'application/msword': None,
    'application/postscript': None,
    'message/rfc822': None,

    # TODO compressed?
    'application/octet-stream': None,
    'application/zip': None,
    'application/x-tar': None,
    'application/gzip': None,
    'application/x-sqlite3': None,
    'application/x-archive': None,
    'application/x-pie-executable': None,
    '.o'  : None,
    'image/jpeg': None,
    '.jpg': None,
    '.png': None,
    'image/png': None,
    '.gif': None,
    '.svg': None,
    '.ico': None,
    'inode/x-empty': None,
    '.class': None,
    '.jar': None,
    '.mp3': None,
    '.mp4': None,
}
# TODO ok, mime doesn't really tell between org/markdown/etc anyway

IGNORE = [
    '.git',
    '.mypy_cache',
    '.pytest_cache',
    'node_modules',
    '__pycache__',
    '.tox',
    '.stack-work',
    # TODO use ripgrep?

    # TODO not sure about these:
    '.gitignore',
    '.babelrc',
]


Replacer = Optional[Callable[[str], str]]

def index(path: Union[List[PathIsh], PathIsh], *, ignored: Union[Sequence[str], str]=(), follow=True, replacer: Replacer=None) -> Iterator[Extraction]:
    # TODO *args?
    # TODO meh, unify with glob traversing..
    paths = path if isinstance(path, list) else [path]
    ignored = (ignored,) if isinstance(ignored, str) else ignored
    opts = Options(
        ignored=ignored,
        follow=follow,
        replacer=replacer,
    )
    for p in paths:
        yield from _index(Path(p), opts=opts)


class Options(NamedTuple):
    ignored: Sequence[str]
    follow: bool
    # TODO option to add ignores? not sure..
    replacer: Replacer


# TODO eh. might be good to use find or fdfind to speed it up...
def _index(path: Path, opts: Options) -> Iterator[Extraction]:
    logger = get_logger()
    pp = path # ugh, for historic reasons..

    if pp.name in IGNORE:
        logger.debug('ignoring %s: default ignore rules', pp)
        return
    if any(fnmatch(str(pp), o) for o in opts.ignored):
        logger.debug('ignoring %s: user ignore rules', pp)
        return

    if pp.is_dir():
        paths = list(pp.glob('*')) # meh
        for p in paths:
            yield from _index(p, opts=opts)
        return

    if pp.is_symlink():
        if opts.follow:
            yield from _index(pp.resolve(), opts=opts)
        else:
            logger.debug('ignoring symlink %s', pp)
        return

    try:
        yield from _index_file(pp, opts=opts)
    except Exception as e:
        # quite likely due to unavoidable race conditions
        yield e


def _index_file(pp: Path, opts: Options) -> Iterator[Extraction]:
    logger = get_logger()
    # TODO use kompress?
    # TODO not even sure if it's used...
    suf = pp.suffix.lower()

    if suf == '.xz': # TODO zstd?
        import lzma
        uname = pp.name[:-len('.xz')]
        uncomp = Path(get_tmpdir().name) / uname
        with lzma.open(pp, 'rb') as cf:
            with uncomp.open('wb') as fb:
                fb.write(cf.read())
        yield from _index(path=uncomp, opts=opts)
        return

    # TODO dispatch org mode here?
    # TODO try/catch?

    if suf not in SMAP:
        pm = mime(pp)
        if pm not in SMAP:
            yield RuntimeError(f"Unexpected file extension: {pp}, {pm}")
            return
        else:
            ip = SMAP.get(pm, None)
        # TODO assume plaintext?
    else:
        ip = SMAP.get(suf, None)
    if ip is None:
        # TODO only log once?
        logger.debug('file type suppressed: %s', pp)
        return

    indexer: Union[Urls, Iterator[Extraction]] = ip(pp) # type: ignore
    # TODO careful, filter out obviously not plaintext? maybe mime could help here??

    fallback_dt = datetime.fromtimestamp(pp.stat().st_mtime, tz=pytz.utc)
    loc = Loc.file(pp)
    replacer = opts.replacer
    for r in indexer:
        if isinstance(r, Exception):
            yield r
            continue
        if isinstance(r, EUrl):
            v = Visit(
                url=r.url,
                dt=fallback_dt,
                locator=loc,
                context='::'.join(r.ctx),
            )
        else:
            v = r
        if replacer is not None:
            upd: Dict[str, Any] = {}
            href = v.locator.href
            if href is not None:
                upd['locator'] = v.locator._replace(href=replacer(href), title=replacer(v.locator.title))
            ctx = v.context
            if ctx is not None:
                # TODO in context, http is unnecessary
                upd['context'] = replacer(ctx)
            v = v._replace(**upd)
        yield v



from pathlib import Path
from typing import Union

PathIsh = Union[Path, str]

def open(path: PathIsh, *args, **kwargs): # TODO is it bytes stream??
    pp = Path(path)
    suf = pp.suffix
    if suf in ('.xz',):
        import lzma
        return lzma.open(pp, *args, **kwargs)
    elif suf in ('.zip',):
        from zipfile import ZipFile
        return ZipFile(pp).open(*args, **kwargs)
    elif suf in ('.lz4',):
        # pylint: disable=import-error
        import lz4.frame # type: ignore
        return lz4.frame.open(str(pp))
    else:
        return pp.open(*args, **kwargs)




# see https://github.com/karlicoss/pymplate for up-to-date reference


from setuptools import setup, find_packages # type: ignore

DEPS_INDEXER = [
    'urlextract',
    # TODO could be optional?
    'python-magic', # for detecting mime types
]

DEPS_SERVER = [
    'tzlocal',
    'hug',
]

def main():
    pkgs = find_packages('src')
    # [pkg] = pkgs
    pkg = 'promnesia' # eh, find some subpackages too?
    setup(
        name=pkg,
        use_scm_version={
            'version_scheme': 'python-simplified-semver',
            'local_scheme': 'dirty-tag',
        },
        setup_requires=['setuptools_scm'],

        zip_safe=False,

        packages=pkgs, # TODO ugh. that's weird. it worked as only ['promnesia'] when installing via PIP ... but not with dev install???
        package_dir={'': 'src'},
        package_data={pkg: ['py.typed']},

        url='https://github.com/karlicoss/promnesia',
        author='Dmitrii Gerasimov',
        author_email='karlicoss@gmail.com',
        description='Enhancement of your browsing history',

        install_requires=[
            *DEPS_INDEXER,
            *DEPS_SERVER,
            'more_itertools',
            'pytz',
            'sqlalchemy', # DB api
            'cachew', # caching with type hints
        ],
        extras_require={
            # TODO make cachew optional?
            # althrough server uses it so not sure...
            'optional': [
                'logzero', # pretty colored logging
            ],
            'testing': [
                 'pytest',
                 'pytest-xdist', # why??

                 'httpie',   # nicer http requests (replace with curl?)
                 'selenium', # browser automations
                 'click',    # confirmations for end2end test (might remove dependency)
            ],
            'linting': [
                 'pytest',
                 'pylint',
                 'mypy',
            ],
            'telegram': [
                'dataset',
            ],
            'reddit': [
                'pyjq',
            ],
            'markdown': [
                'mistletoe',
            ],
            'org': [
                'orgparse',
            ],
            'html': [
                'beautifulsoup4', # extracting links from the page
                'lxml', # bs4 backend
            ],
            'HPI': [
                'HPI', # pypi version
                # 'HPI @ git+https://github.com/karlicoss/hpi.git', # uncomment to test against github version (useful for one-off CI run)
                # 'HPI @ git+file://DUMMY/path/to/local/hpi'    , # uncomment to test against version on the disc
            ],
            # TODO make 'all' group?
        },
        entry_points={
            'console_scripts': ['promnesia=promnesia.__main__:main'],
        }
    )


if __name__ == "__main__":
    main()



from itertools import groupby

from promnesia.sources import auto

from common import tdata

sa2464 = 'https://www.scottaaronson.com/blog/?p=2464'

_JSON_URLS = {
    'https://johncarlosbaez.wordpress.com/2016/09/09/struggles-with-the-continuum-part-2/',
    sa2464,
}


def makemap(visits):
    key = lambda v: v.url
    def it():
        for k, g in groupby(sorted(visits, key=key), key=key):
            yield k, list(sorted(g))
    return dict(it())


def test_json():
    mm = makemap(auto.index(tdata('auto/pocket.json')))
    assert mm.keys() == _JSON_URLS

    # TODO not sure if they deserve separate visits..
    [v1, v2] = mm[sa2464]
    assert v1.context == 'list::yyy::given_url'
    assert v1.locator.href.startswith('emacs:')
    assert v1.locator.href.endswith('pocket.json')
    # TODO line number?


def test_auto():
    mm = makemap(auto.index(tdata('auto')))
    org_link = 'https://www.youtube.com/watch?v=rHIkrotSwcc'
    assert {
        *_JSON_URLS,
        org_link,
    }.issubset(mm.keys())

    [v] = mm[org_link]
    assert v.locator.href.endswith('file.org') # TODO link number or jump to heading?
    assert "xxx /r/cpp\n I've enjoyed [Chandler Carruth's" in v.context



'''
A more sophisticated example of config for Promnesia
'''

from pathlib import Path

OUTPUT_DIR = Path('/data/promnesia')

from promnesia import Source

# 'auto' indexer tries its best at indexing plaintext stuff
# - plaintext like org-mode/markdown/HTML
# - structured formats like JSON and CSV
from promnesia.sources import auto

# 'guess' indexer can do even more in addition:
# - HTTP links (to index the contents of a website)
# - Github links (to index the contents of a git repository
from promnesia.sources import guess
# TODO there is a very thin link between 'auto' and 'guess'... I mught merge them in the future?


# there are also
# this is an incomplete list, just the (perhaps) most interesting ones
from promnesia.sources import telegram
from promnesia.sources import takeout, instapaper, pocket, fbmessenger, twitter, roamresearch, hypothesis

# now we can specify the sources
# this is a required setting
SOURCES = [
    # handle my knowledge base: extract links from Org-mode and Markdown files
    Source(
        auto.index,
        '/path/to/my/blog',

        # you can specify optional name, so you can see where the URL is coming from in the extension
        name='blog',
    ),

    # we can index another path under a different name (for convenience)
    Source(
        auto.index,
        '/path/to/personal/notes',
        ignored=['*.html'],  # we can exclude certain files if we want

        name='notes',
    ),


    # will clone the repository and index its contents!
    Source(
        guess.index, # you could also use 'vcs.index', but auto can handle it anyway
        'https://github.com/karlicoss/exobrain',
        name='exobrain',
    ),


    # Uses the output of telegram_backup tool: https://github.com/fabianonline/telegram_backup#usage
    Source(
        telegram.index,
        '/data/telegram/database.sqlite',
        name='tg',
    ),

    ####
    # the following modules are relying on HPI package (https://github.com/karlicoss/HPI#whats-inside) for the data
    # see setup guide: https://github.com/karlicoss/HPI/blob/master/doc/SETUP.org
    # and 'Usage examples': https://github.com/karlicoss/HPI/blob/master/doc/SETUP.org#usage-examples
    Source(takeout.index),
    Source(instapaper.index),
    Source(pocket.index),
    Source(fbmessenger.index),
    Source(twitter.index),
    Source(roamresearch.index),

    # you can also use a lambda
    # sometimes it's useful for config hacking/avoiding early imports
    lambda: Source(hypothesis.index),
]




'''
Optional setting.
A directory to keep intemediate caches in order to speed up indexing.
'''
CACHE_DIR = '/tmp/promnesia_cache/'


'''
Optional setting.
Yu can specify the URLs you don't want to be indexed.
You might want it because there are too many of them, or for security/privacy reasons.
'''
FILTERS = [
    'mail.google.com',
    'web.telegram.org/#/im',
    'vk.com/im',
    '192.168.0.',

    # you can use regexes too!
    'redditmedia.com.*.(jpg|png|gif)',

]



import importlib
import sys
from contextlib import contextmanager


@contextmanager
def extra_path(path):
    # TODO hmm, tmp_path would be a better name, but then it can be confused with FS path
    ps = str(path)
    if ps in sys.path:
        yield # no need to change anything
    else:
        sys.path.insert(0, ps)
        try:
            yield
        finally:
            sys.path.remove(ps)


# TODO and now it raises the question if it's necessary at all.. probably not so useful??
def import_from(path, name: str, package=None):
    with extra_path(path):
        return importlib.import_module(name, package=package)


def test_import_from(tmp_path):
    from pathlib import Path
    tp = Path(tmp_path)
    tm = tp / 'testmodule'
    tm.mkdir()
    tf = tm / 'testfile.py'
    tf.write_text('''
class AAA:
    passed = "PASSED"
''')
    tm = import_from(tp, 'testmodule.testfile')
    assert tm.AAA.passed == "PASSED" # type: ignore[attr-defined]



from pathlib import Path
from subprocess import check_call, run
from typing import Set, Dict, Optional, Union, Sequence, Tuple

from indexer_test import populate_db
from common import tdir, under_ci

import pytest

import platform
system = platform.system()

# TODO move to common?
ROOT = Path(__file__).absolute().parent.parent

testdata = ROOT / 'testdata'


def index(cfg: Path):
    from promnesia.__main__ import do_index
    do_index(cfg)


def test_example_config(tdir):
    example = ROOT / 'config.py.example'
    ex = example.read_text()
    if under_ci():
        # TODO ugh fucking hell I couldn't find a single path that has HTMLs both on macos and ubuntu
        # and using /usr/share/docs locally might index a bit too much
        if system == 'Darwin':
            repl = '/usr/share/doc/cups/'
        else:
            repl = '/usr/share/doc/python3/'
        ex = ex.replace('/usr/share/doc/python3/html/faq', repl)
    cfg = tdir / 'test_config.py'
    cfg.write_text(ex)
    index(cfg)


# TODO not sure if makes a lot of sense? maybe on no indexers should actually error
def test_empty(tdir):
    cfg = tdir / 'test_config.py'
    cfg.write_text(f"""
SOURCES = []
OUTPUT_DIR = '{tdir}'
    """)
    index(cfg)


from sqlalchemy import create_engine, MetaData, exists # type: ignore
from sqlalchemy import Column, Table # type: ignore
from cachew import NTBinder
from promnesia.common import DbVisit # TODO ugh. figure out pythonpath


def _get_stuff(outdir: Path):
    db_path = outdir / 'promnesia.sqlite'
    assert db_path.exists()

    engine = create_engine(f'sqlite:///{db_path}')

    binder = NTBinder.make(DbVisit)

    meta = MetaData(engine)
    table = Table('visits', meta, *binder.columns)

    return engine, binder, table

# TODO a bit shit... why did I make it dict at first??
Urls = Union[
              Dict[str, Optional[str]],
    Sequence[Tuple[str, Optional[str]]],
]

def index_urls(urls: Urls):
    uuu = list(urls.items()) if isinstance(urls, dict) else urls

    def idx(tdir: Path):
        cfg = tdir / 'test_config.py'
        cfg.write_text(f"""
OUTPUT_DIR = '{tdir}'

from promnesia.common import Source, PreVisit, Loc
from datetime import datetime, timedelta
indexer = Source(
    lambda: [PreVisit(
        url=url,
        dt=datetime.min + timedelta(days=5000) + timedelta(hours=i),
        locator=Loc.make('test'),
        context=ctx,
    ) for i, (url, ctx) in enumerate({uuu})],
    name='test',
)

SOURCES = [indexer]
""")
        index(cfg)
    return idx


def index_hypothesis(tdir: Path):
    # TODO meh..
    hypexport_path = Path(tdir) / 'hypexport'
    check_call([
        'git',
        'clone',
        # TODO do I need --recursive here?? I guess it's only for fetching new data?
        'https://github.com/karlicoss/hypexport',
        str(hypexport_path),
    ])

    cfg = tdir / 'test_config.py'
    # TODO ok, need to simplify this...
    cfg.write_text(f"""
OUTPUT_DIR = '{tdir}'

from promnesia.common import Indexer as I

def hyp_extractor():
    import my.config
    class hypothesis:
        export_path = '{testdata}/hypothesis/netrights-dashboards-mockup/data/annotations.json'
    my.config.hypothesis = hypothesis

    # TODO good example for documentaion
    from my.cfg import set_repo
    set_repo('hypexport', '{hypexport_path}')

    import promnesia.sources.hypothesis as hypi
    return I(
        hypi.index,
        src='hyp',
    )

# in addition, test for lazy indexers. useful for importing packages
SOURCES = [hyp_extractor]
    """)
    index(cfg)


def index_local_chrome(tdir: Path):
    # TODO mm, would be good to keep that for proper end2end
    # inp = Path('/L/data/promnesia/testdata/chrome-history/History') # TODO make it accessible to the repository
    # merged = tdir / 'chrome-merged.sqlite'
    # populate_db.merge_from('chrome', from_=inp, to=merged)

    merged = Path('/L/data/promnesia/testdata/chrome.sqlite')

    cfg = tdir / 'test_config.py'
    cfg.write_text(f"""
OUTPUT_DIR = '{tdir}'

from promnesia.common import Indexer as I
from promnesia.sources.browser import chrome

chrome_extractor = I(chrome, '{merged}', src='chrome')

SOURCES = [chrome_extractor]
""")
    index(cfg)


def test_hypothesis(tdir):
    index_hypothesis(tdir)

    # TODO copy pasting from server; need to unify
    engine, binder, table = _get_stuff(tdir)
    query = table.select()
    with engine.connect() as conn:
        visits = [binder.from_row(row) for row in conn.execute(query)]

    assert len(visits) > 100

    [vis] = [x for x in visits if 'fundamental fact of evolution' in x.context]

    assert vis.norm_url == 'wired.com/2017/04/the-myth-of-a-superhuman-ai'
    assert vis.orig_url == 'https://www.wired.com/2017/04/the-myth-of-a-superhuman-ai/'
    assert vis.locator.href == 'https://hyp.is/_Z9ccmVZEeexBOO7mToqdg/www.wired.com/2017/04/the-myth-of-a-superhuman-ai/'
    assert 'misconception about evolution is fueling misconception about AI' in vis.context # contains notes as well


def test_comparison(tdir: Path):
    from promnesia.compare import compare_files

    idx = index_urls({
        'https://example.com': None,
        'https://en.wikipedia.org/wiki/Saturn_V': None,
        'https://plato.stanford.edu/entries/qualia': None,
    })
    idx(tdir)
    db     = tdir / 'promnesia.sqlite'
    old_db = tdir / 'promnesia-old.sqlite'
    import shutil
    shutil.move(db, old_db)

    idx2 = index_urls({
        'https://example.com': None,
        'https://www.reddit.com/r/explainlikeimfive/comments/1ev6e0/eli5entropy': None,
        'https://en.wikipedia.org/wiki/Saturn_V': None,
        'https://plato.stanford.edu/entries/qualia': None,
    })
    idx2(tdir)

    # should not crash, as there are more links in the new database
    assert len(list(compare_files(old_db, db))) == 0

    assert len(list(compare_files(db, old_db))) == 1




#!/usr/bin/python3
__package__ = 'promnesia'  # ugh. hacky way to make hug work properly...

import os
import json
from datetime import timedelta, datetime
from pathlib import Path
import logging
from functools import lru_cache
from typing import Collection, List, NamedTuple, Dict, Optional


from cachew import NTBinder

import pytz
from pytz.tzinfo import BaseTzInfo # type: ignore

import hug # type: ignore
import hug.types as T # type: ignore

from sqlalchemy import create_engine, MetaData, exists, literal, between, or_, and_ # type: ignore
from sqlalchemy import Column, Table, func # type: ignore


from .common import PathWithMtime, DbVisit, Url, Loc, setup_logger, PathIsh
from .cannon import canonify

_ENV_CONFIG = 'PROMNESIA_CONFIG'


# TODO not sure about utc in database... keep orig timezone?

# meh. need this since I don't have hooks in hug to initialize logging properly..
@lru_cache(1)
def get_logger():
    logger = logging.getLogger('promnesia')
    setup_logger(logger, level=logging.DEBUG)
    return logger


def get_version() -> str:
    from pkg_resources import get_distribution
    return get_distribution('promnesia').version


class ServerConfig(NamedTuple):
    db: Path
    timezone: BaseTzInfo

    @classmethod
    def make(cls, db: PathIsh, timezone: str) -> 'ServerConfig':
        tz = pytz.timezone(timezone)
        return cls(db=Path(db), timezone=tz)


@lru_cache(1)
def get_config() -> ServerConfig:
    cfg = os.environ.get(_ENV_CONFIG)
    assert cfg is not None
    return ServerConfig.make(**json.loads(cfg))


# TODO use that?? https://github.com/timothycrosley/hug/blob/develop/tests/test_async.py

# TODO how to return exception in error?

def as_json(v: DbVisit) -> Dict:
    # yep, this is NOT %Y-%m-%d as is seems to be the only format with timezone that Date.parse in JS accepts. Just forget it.
    dts = v.dt.strftime('%d %b %Y %H:%M:%S %z')
    loc = v.locator
    # TODO is locator always present??
    return {
        # TODO do not display year if it's current year??
        'dt': dts,
        # TODO the frontend had some bug with handling empty string as src. fix that later
        'src': v.src or 'unnamed',
        'context': v.context,
        'duration': v.duration,
        'locator': {
            'title': loc.title,
            'href' : loc.href,
        },
        'original_url'  : v.orig_url,
        'normalised_url': v.norm_url,
    }


def get_db_path() -> Path:
    config = get_config()
    db = config.db
    assert db.exists(), db
    return db


# TODO maybe, keep db connection? need to recycle it properly..
@lru_cache(1)
# PathWithMtime aids lru_cache in reloading the sqlalchemy binder
def _get_stuff(db_path: PathWithMtime):
    get_logger().info('Reloading DB: %s', db_path)
    # TODO how to open read only?
    engine = create_engine(f'sqlite:///{db_path.path}') # , echo=True)

    binder = NTBinder.make(DbVisit)

    meta = MetaData(engine)
    table = Table('visits', meta, *binder.columns)

    return engine, binder, table


def get_stuff(db_path=None): # TODO better name
    # ok, it will always load from the same db file; but intermediate would be kinda an optional dump.
    if db_path is None:
        db_path = get_db_path()
    return _get_stuff(PathWithMtime.make(db_path))



def search_common(url: str, where):
    logger = get_logger()
    config = get_config()

    logger.info('url: %s', url)
    original_url = url
    url = canonify(url)
    logger.info('normalised url: %s', url)

    engine, binder, table = get_stuff()

    query = table.select().where(where(table=table, url=url))

    logger.info('query: %s', query)

    with engine.connect() as conn:
        visits = [binder.from_row(row) for row in conn.execute(query)]

    logger.debug('got %d visits from db', len(visits))

    vlist = []
    for vis in visits:
        dt = vis.dt
        if dt.tzinfo is None:
            tz = config.timezone
            dt = tz.localize(dt)
            vis = vis._replace(dt=dt)
        vlist.append(vis)

    logger.debug('responding with %d visits', len(vlist))
    # TODO respond with normalised result, then frontent could choose how to present children/siblings/whatever?
    return {
        'orginal_url'   : original_url,
        'normalised_url': url,
        'visits': list(map(as_json, vlist)),
    }


@hug.local()
@hug.post('/status')
def status():
    '''
    Ideally, status will always respond, regardless the internal state of the backend?
    '''
    # TODO hug stats?

    db_path: Optional[str]
    try:
        db_path = str(get_db_path())
        # TODO use 'db_stats' instead? add count or something else
    except Exception as e:
        # TODO not sure how to properly communicate the error to frontend?
        db_path = None

    version: Optional[str]
    try:
        version = get_version()
    except Exception as e:
        version = None

    return {
        'db'     : db_path,
        'version': version,
    }
# TODO might be good to include the frontend version in the requests?


@hug.local()
@hug.post('/visits')
def visits(
        url: T.text,
):
    return search_common(
        url=url,
        # odd, doesn't work just with: x or (y and z)
        where=lambda table, url: or_(
            table.c.norm_url == url,
            and_(table.c.context != None, table.c.norm_url.startswith(url, autoescape=True))
        ),
    )


@hug.local()
@hug.post('/search')
def search(
        url: T.text
):
    # TODO rely on hug logger for query
    return search_common(
        url=url,
        where=lambda table, url: or_(
            table.c.norm_url.contains(url, autoescape=True),
            # TODO hmm. think about it, not sure if I need proper indexer for fuzzy search etc?
            table.c.context.contains(url, autoescape=True),
        ),
    )


@hug.local()
@hug.post('/search_around')
def search_around(
        timestamp: T.number,
):
    utc_timestamp = timestamp # old 'timestamp' name is legacy

    # TODO meh. use count instead?
    delta_back  = timedelta(hours=3).total_seconds()
    delta_front = timedelta(minutes=2).total_seconds()
    # TODO not sure about front.. but it also serves as quick hack to accomodate for all the truncations etc
    return search_common(
        url='http://dummy.org', # TODO remove it from search_common
        where=lambda table, url: between(
            # %s is a unix timestamp
            # TODO careful.. not sure how datetime works w.r.t. datetime string without the timezone info..
            func.strftime('%s', func.datetime(table.c.dt)) - literal(utc_timestamp),
            literal(-delta_back),
            literal(delta_front),
        ),
    )

@hug.local()
@hug.post('/visited')
def visited(
        urls, # TODO type
):
    logger = get_logger()

    logger.debug(urls)
    norms = [(u, canonify(u)) for u in urls]
    # logger.debug('\n'.join(f'{u} -> {nu}' for u, nu in norms))

    nurls = [n[1] for n in norms]
    engine, binder, table = get_stuff()

    snurls = list(sorted(set(nurls)))
    # sqlalchemy doesn't seem to support SELECT FROM (VALUES (...)) in its api
    # also doesn't support array binding...
    # https://stackoverflow.com/questions/13190392/how-can-i-bind-a-list-to-a-parameter-in-a-custom-query-in-sqlalchemy
    bstring = ','.join(f'(:b{i})'   for i, _ in enumerate(snurls))
    bdict = {            f'b{i}': v for i, v in enumerate(snurls)}

    query = f"""
WITH cte(queried) AS (SELECT * FROM (values {bstring}))
SELECT queried
    FROM cte JOIN visits
    ON queried = visits.norm_url
    """
    # hmm that was quite slow...
    # SELECT queried FROM cte WHERE EXISTS (SELECT 1 FROM visits WHERE queried = visits.norm_url)
    logger.debug(bdict)
    logger.debug(query)
    with engine.connect() as conn:
        res = list(conn.execute(query, bdict))
        present = {x[0] for x in res}
    results = [nu in present for nu in nurls]

    # logger.debug('\n'.join(
    #     f'{"X" if v else "-"} {u} -> {nu}' for v, (u, nu) in zip(results, norms)
    # ))
    return results


def _run(*, port: str, db: Path, timezone: str, quiet: bool):
    logger = get_logger()
    env = {
        **os.environ,
        # not sure if there is a simpler way to communicate with hug..
        _ENV_CONFIG: json.dumps({'db': str(db), 'timezone': timezone}),
    }
    args = [
        'python3',
        '-m', 'hug', # TODO eh, not sure about this. what if user had it already installed?? it's a mess..
        *(['--silent'] if quiet else []),
        '-p', port,
        '-f', __file__,
    ]
    logger.info('Running server: %s', args)
    os.execvpe('python3', args, env)


def run(args):
    _run(port=args.port, db=args.db, timezone=args.timezone, quiet=args.quiet)


_DEFAULT_CONFIG = Path('config.py')


def get_system_tz() -> str:
    logger = get_logger()
    try:
        import tzlocal # type: ignore
        return tzlocal.get_localzone().zone
    except Exception as e:
        logger.exception(e)
        logger.error("Couldn't determine system timezone. Falling back to UTC. Please report this as a bug!")
        return 'UTC'


# TODO rename to 'backend'?
def setup_parser(p):
    p.add_argument('--port'    , type=str , default='13131', help='Port for communicating with extension')
    # TODO mm. should add fallback timezone to frontend instead I guess?
    p.add_argument('--db'      , type=Path, required=True  , help='Path to the link database (required)')
    p.add_argument('--timezone', type=str , default=get_system_tz(), help='Fallback timezone, defaults to the system timezone if not specified')
    p.add_argument('--quiet'              , action='store_true', help='Less logging')




from pathlib import Path
from typing import List, Optional, Union, NamedTuple
import importlib.util
import warnings

import pytz

from .common import PathIsh, get_tmpdir


class Config(NamedTuple):
    OUTPUT_DIR: PathIsh
    # TODO remove default from sources once migrated
    SOURCES: List = []
    INDEXERS: List = []
    CACHE_DIR: Optional[PathIsh] = None
    FILTERS: List[str] = []

    @property
    def sources(self):
        if self.INDEXERS is not None:
            warnings.warn("'INDEXERS' is deprecated. Please use 'SOURCES'!", DeprecationWarning)

        res = self.SOURCES or self.INDEXERS
        # TODO enable it?
        # assert len(res) > 0, "Expected some sources"
        return res

    @property
    def cache_dir(self) -> Path:
        cd = self.CACHE_DIR
        # TODO maybe do not use cache if it's none?
        assert cd is not None
        res = Path(cd)
        res.mkdir(exist_ok=True) # TODO not sure about parents=True
        return res

    @property
    def output_dir(self) -> Path:
        return Path(self.OUTPUT_DIR)


instance: Optional[Config] = None


def has() -> bool:
    return instance is not None

def get() -> Config:
    assert instance is not None
    return instance


def load_from(config_file: Path) -> None:
    global instance
    instance = import_config(config_file)


def reset() -> None:
    global instance
    assert instance is not None
    instance = None


def import_config(config_file: PathIsh) -> Config:
    p = Path(config_file)

    # TODO just exec??
    name = p.stem
    spec = importlib.util.spec_from_file_location(name, p) # type: ignore
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod) # type: ignore

    d = {}
    for f in Config._fields:
        if hasattr(mod, f):
            d[f] = getattr(mod, f)
    return Config(**d)



from typing import Iterator, Optional

from ..common import Extraction, get_logger, Visit, Loc, PathIsh

import my.instapaper as ip


def index() -> Iterator[Extraction]:
    logger = get_logger()

    for p in ip.pages():
        bm = p.bookmark
        hls = p.highlights

        def visit(**kwargs):
            return Visit(
                url=bm.url,
                **kwargs,
            )

        if len(hls) == 0:
            yield visit(
                dt=bm.dt,
                context=None,
                locator=Loc.make(title='instapaper', href=bm.instapaper_link),
            )
        else:
            for hl in p.highlights:
                cparts = [hl.text]
                if hl.note is not None:
                    cparts.append('comment: ' + hl.note)
                yield visit(
                    dt=hl.dt,
                    context='\n'.join(cparts),
                    locator=Loc.make(title='instapaper', href=hl.instapaper_link),
                )




#!/usr/bin/env python3
from contextlib import contextmanager
from pathlib import Path
from datetime import datetime
from tempfile import TemporaryDirectory
from subprocess import check_call, check_output
from time import sleep
from typing import NamedTuple, Optional

import pytest # type: ignore

if __name__ == '__main__':
    # TODO ugh need to figure out PATH
    # python3 -m pytest -s tests/server_test.py::test_query
    pytest.main(['-s', __file__])


from selenium import webdriver # type: ignore
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.by import By


from common import under_ci, skip_if_ci, uses_x
from integration_test import index_hypothesis, index_local_chrome, index_urls
from server_test import wserver
from firefox_helper import open_extension_page


class Browser(NamedTuple):
    dist: str
    headless: bool

    @property
    def name(self) -> str:
        return self.dist.split('-')[0] # TODO meh

    def skip_ci_x(self):
        if under_ci() and not self.headless:
            pytest.skip("Only can't use headless browser on CI")


FF  = Browser('firefox', headless=False)
CH  = Browser('chrome' , headless=False)
FFH = Browser('firefox', headless=True)
# NOTE: sadly headless chrome doesn't support extensions..
# https://stackoverflow.com/a/45372648/706389
# there is some workaround, but it's somewhat tricky...
# https://stackoverflow.com/a/46475980/706389


# TODO ugh, I guess it's not that easy to make it work because of isAndroid checks...
# I guess easy way to test if you really want is to temporary force isAndroid to return true in extension...
FM  = Browser('firefox-mobile', headless=False)


def browser_(driver):
    name = driver.name
    # TODO figure out headless??
    if name == 'firefox':
        return FF
    elif name == 'chrome':
        return CH
    else:
        raise AssertionError(driver)


def get_addon_path(kind: str) -> Path:
    # TODO compile first?
    addon_path = (Path(__file__).parent.parent / 'extension' / 'dist' / kind).absolute()
    assert addon_path.exists()
    assert (addon_path / 'manifest.json').exists()
    return addon_path


def get_hotkey(driver, cmd: str) -> str:
    # TODO shit, need to unify this...
    if driver.name == 'chrome':
        chrome_profile = Path(driver.capabilities['chrome']['userDataDir'])
        prefs_file = chrome_profile / 'Default/Preferences'
        import json
        prefs = json.loads(prefs_file.read_text())
        # "commands": {
        #        "linux:Ctrl+Shift+V": {
        #            "command_name": "show_dots",
        #            "extension": "ceedkmkoeooncekjljapnkkjhldddcid",
        #            "global": false
        #        }
        #    },
        cmd_map = {cmd['command_name']: k.split(':')[-1] for k, cmd in prefs['extensions']['commands'].items()}
    else:
        # TODO remove hardcoding somehow...
        # perhaps should be extracted somewhere..
        cmd_map = {
            Command.SHOW_DOTS        : 'Ctrl+Alt+v',
            '_execute_browser_action': 'Ctrl+Alt+e',
            'search'                 : 'Ctrl+Alt+h',
        }
    return cmd_map[cmd].split('+')


def _get_webdriver(tdir: Path, browser: Browser, extension=True):
    addon = get_addon_path(kind=browser.dist)
    if browser.name == 'firefox':
        profile = webdriver.FirefoxProfile(str(tdir))
        options = webdriver.FirefoxOptions()
        options.headless = browser.headless
        # use firefox from here to test https://www.mozilla.org/en-GB/firefox/developer/
        driver = webdriver.Firefox(profile, options=options)

        # driver = webdriver.Firefox(profile, firefox_binary='/L/soft/firefox-dev/firefox/firefox', options=options)
        # TODO how to pass it here properly?

        if extension:
            driver.install_addon(str(addon), temporary=True)
    elif browser.name == 'chrome':
        # TODO ugh. very hacky...
        assert extension, "TODO add support for extension arg"
        ex = tdir / 'extension.zip'
        files = [x.name for x in addon.iterdir()]
        check_call(['apack', '-q', str(ex), *files], cwd=addon)
        # looks like chrome uses temporary dir for data anyway
        options = webdriver.ChromeOptions()
        options.headless = browser.headless
        options.add_extension(ex)
        driver = webdriver.Chrome(options=options)
    else:
        raise RuntimeError(f'Unexpected browser {browser}')
    return driver


# TODO copy paste from grasp
@contextmanager
def get_webdriver(browser: Browser, extension=True):
    with TemporaryDirectory() as td:
        tdir = Path(td)
        driver = _get_webdriver(tdir, browser=browser, extension=extension)
        try:
            yield driver
        finally:
            driver.quit()


def set_host(*, driver, host: str, port: str):
    ep = driver.find_element_by_id('host_id') # TODO rename to 'backend'?
    ep.clear()
    ep.send_keys(f'{host}:{port}')


def save_settings(driver):
    se = driver.find_element_by_id('save_id')
    se.click()

    driver.switch_to.alert.accept()


LOCALHOST = 'http://localhost'


def configure(
        driver,
        *,
        host: Optional[str]=LOCALHOST,
        port: Optional[str]=None,
        show_dots: bool=True,
        highlights: Optional[bool]=None,
        blacklist=None,
        notification: Optional[bool]=None,
        position: Optional[str]=None,
):
    def set_checkbox(cid: str, value: bool):
        cb = driver.find_element_by_id(cid)
        selected = cb.is_selected()
        if selected != value:
            cb.click()


    # TODO log properly
    print(f"Setting: port {port}, show_dots {show_dots}")

    open_extension_page(driver, page='options_page.html')
    sleep(1) # err, wtf? otherwise not always interacts with the elements correctly

    assert not (host is None) ^ (port is None), (host, port)
    if host is not None:
        assert port is not None
        set_host(driver=driver, host=host, port=port)

    # dots = driver.find_element_by_id('dots_id')
    # if dots.is_selected() != show_dots:
    #     dots.click()
    # assert dots.is_selected() == show_dots
    set_checkbox('verbose_errors_id', False)

    if highlights is not None:
        set_checkbox('highlight_id', highlights)

    if notification is not None:
        set_checkbox('contexts_popup_id', notification)

    if position is not None:
        set_position(driver, position)

    if blacklist is not None:
        bl = driver.find_element_by_id('blacklist_id') # .find_element_by_tag_name('textarea')
        bl.click()
        # ugh, that's hacky. presumably due to using Codemirror?
        bla = driver.switch_to_active_element()
        bla.send_keys('\n'.join(blacklist))

    save_settings(driver)


# legacy name
configure_extension = configure


def get_window_id(driver):
    if driver.name == 'firefox':
        pid = str(driver.capabilities['moz:processID'])
    else:
        # ugh nothing in capabilities...
        pid = check_output(['pgrep', '-f', 'chrome.*enable-automation']).decode('utf8').strip()
    # https://askubuntu.com/a/385037/427470
    return get_wid_by_pid(pid)


def get_wid_by_pid(pid: str):
    wids = check_output(['xdotool', 'search', '--pid', pid]).decode('utf8').splitlines()
    wids = [w.strip() for w in wids if len(w.strip()) > 0]

    def has_wm_desktop(wid: str) -> bool:
        # TODO no idea why is that important. found out experimentally
        out = check_output(['xprop', '-id', wid, '_NET_WM_DESKTOP']).decode('utf8')
        return 'not found' not in out

    [wid] = filter(has_wm_desktop, wids)
    return wid


def focus_browser_window(driver):
    wid = get_window_id(driver)
    check_call(['xdotool', 'windowactivate', wid])


def trigger_callback(driver, callback):
    focus_browser_window(driver)
    sleep(0.5)
    callback()


def send_key(key):
    if isinstance(key, str):
        key = key.split('+')

    print(f"sending hotkey! {key}")
    import pyautogui # type: ignore
    pyautogui.hotkey(*key)


def trigger_hotkey(driver, hotkey):
    trigger_callback(driver, lambda: send_key(hotkey))


def trigger_command(driver, cmd):
    trigger_hotkey(driver, get_hotkey(driver, cmd))


class TestHelper(NamedTuple):
    driver: webdriver.Remote

    def open_page(self, page: str) -> None:
        open_extension_page(self.driver, page)

    def move_to(self, element):
        from selenium.webdriver.common.action_chains import ActionChains # type: ignore
        ActionChains(self.driver).move_to_element(element).perform()

    def switch_to_sidebar(self):
        self.driver.switch_to.default_content()
        self.driver.switch_to.frame('promnesia-sidebar')

    def command(self, cmd):
        trigger_command(self.driver, cmd)

    def activate(self):
        self.command(Command.ACTIVATE)

    def mark_visited(self):
        self.command(Command.SHOW_DOTS)

    def wid(self) -> str:
        return get_window_id(self.driver)

    def screenshot(self, path):
        # ugh, webdriver's save_screenshot doesn't behave well with frames
        check_call(['import', '-window', self.wid(), path])


def confirm(what: str):
    import click # type: ignore
    click.confirm(what, abort=True)


@contextmanager
def _test_helper(tmp_path, indexer, test_url: Optional[str], show_dots: bool=False, browser: Browser=FFH):
    tdir = Path(tmp_path)

    indexer(tdir)
    with wserver(db=tdir / 'promnesia.sqlite') as srv, get_webdriver(browser=browser) as driver:
        port = srv.port
        configure_extension(driver, port=port, show_dots=show_dots)
        sleep(0.5)

        if test_url is not None:
            driver.get(test_url)
            sleep(3)
        else:
            driver.get('about:blank')

        yield TestHelper(driver=driver)

class Command:
    SHOW_DOTS = 'mark_visited'
    ACTIVATE  = '_execute_browser_action'
    SEARCH    = 'search'
# TODO assert this against manifest?


def browsers(*br):
    from functools import wraps
    def dec(f):
        @pytest.mark.with_browser
        @pytest.mark.parametrize('browser', br, ids=lambda b: b.dist.replace('-', '_') + ('_headless' if b.headless else ''))
        @wraps(f)
        def ff(*args, **kwargs):
            return f(*args, **kwargs)
        return ff
    return dec


PYTHON_DOC_URL = 'file:///usr/share/doc/python3/html/index.html'


@browsers(FFH, CH)
def test_installs(tmp_path, browser):
    browser.skip_ci_x()

    with get_webdriver(browser=browser):
        # just shouldn't crash
        pass


@browsers(FFH, CH)
def test_settings(tmp_path, browser):
    browser.skip_ci_x()

    # TODO fixture for driver?
    with get_webdriver(browser=browser) as driver:
        configure_extension(driver, port='12345', show_dots=False)
        driver.get('about:blank')
        open_extension_page(driver, page='options_page.html')
        hh = driver.find_element_by_id('host_id')
        assert hh.get_attribute('value') == 'http://localhost:12345'


@browsers(FFH, FF, CH)
def test_backend_status(tmp_path, browser):
    browser.skip_ci_x()

    with get_webdriver(browser=browser) as driver:
        open_extension_page(driver, page='options_page.html')
        sleep(1) # ugh. for some reason pause here seems necessary..
        set_host(driver=driver, host='https://nosuchhost.com', port='1234')
        driver.find_element_by_id('backend_status_id').click()
        sleep(1 + 0.5) # needs enough time for timeout to trigger...

        alert = driver.switch_to.alert
        assert 'ERROR' in alert.text

        # TODO ugh, not sure why it just stucks on headless...
        # perhaps use policy instead? https://stackoverflow.com/a/43946973/706389
        if not browser.headless:
            driver.switch_to.alert.accept()
            sleep(0.5)
            # ugh. extra alert...
            driver.switch_to.alert.accept()

        # TODO implement positive check??

def set_position(driver, settings: str):
    browser = browser_(driver)

    # TODO figure out browser from driver??
    field = driver.find_element_by_xpath('//*[@id="position_css_id"]')
    if browser.name == 'chrome':
        # ugh... for some reason wouldn't send the keys...
        field.click()
        import pyautogui # type: ignore
        pyautogui.press(['backspace'] * 100 + ['delete'] * 100)
        pyautogui.typewrite(settings, interval=0.1)
    else:
        area = field.find_element_by_xpath('.//textarea')
        area.send_keys([Keys.DELETE] * 500)
        area.send_keys(settings)



@browsers(FF, CH)
def test_sidebar_bottom(browser):
    browser.skip_ci_x()

    with get_webdriver(browser=browser) as driver:
        open_extension_page(driver, page='options_page.html')
        sleep(1) # ugh. for some reason pause here seems necessary..

        # for some reason area.clear() caused
        # selenium.common.exceptions.ElementNotInteractableException: Message: Element <textarea> could not be scrolled into view

        settings = """
#promnesia-sidebar {
    --bottom: 1;
    --size: 20%;
}"""

        set_position(driver, settings)

        save_settings(driver)

        driver.get(PYTHON_DOC_URL)

        trigger_command(driver, Command.ACTIVATE)
        confirm("You should see sidebar below")


@uses_x
@browsers(FF, CH)
def test_blacklist_custom(tmp_path, browser):
    # TODO make confirm a fixture? so we can automatically skip them on ci
    with get_webdriver(browser=browser) as driver:
        configure_extension(driver, port='12345', blacklist=('stackoverflow.com',))
        driver.get('http://stackoverflow.com')
        trigger_command(driver, Command.ACTIVATE)
        confirm('page should be blacklisted (black icon), your should see an error notification')


@uses_x
@browsers(FF, CH)
def test_blacklist_builtin(tmp_path, browser):
    with get_webdriver(browser=browser) as driver:
        configure_extension(driver, port='12345')
        driver.get('https://www.hsbc.co.uk/mortgages/')
        confirm('page should be blacklisted (black icon)')


@uses_x
@browsers(FF, CH)
def test_add_to_blacklist(tmp_path, browser):
    with get_webdriver(browser=browser) as driver:
        configure_extension(driver, port='12345')
        driver.get('https://example.com')
        chain = webdriver.ActionChains(driver)
        chain.move_to_element(driver.find_element_by_tag_name('h1')).context_click().perform()

        # looks like selenium can't interact with browser context menu...
        import pyautogui # type: ignore

        if driver.name == 'chrome':
            offset = 2 # Inspect, View page source
        else:
            offset = 0
        pyautogui.typewrite(['up'] + ['up'] * offset + ['enter'], interval=0.5)

        confirm('shows prompt with alert to enter pattern to block?')
        driver.switch_to.alert.accept()
        # ugh, seems necessary to guard with sleep; otherwise racey
        sleep(0.5)

        driver.get(driver.current_url)
        confirm('page should be blacklisted (black icon)')


@uses_x
@browsers(FF, CH)
def test_visits(tmp_path, browser):
    test_url = "http://www.e-flux.com/journal/53/59883/the-black-stack/"
    # test_url = "file:///usr/share/doc/python3/html/library/contextlib.html" # TODO ??
    with _test_helper(tmp_path, index_hypothesis, test_url, browser=browser) as helper:
        trigger_command(helper.driver, Command.ACTIVATE)
        confirm('you should see hypothesis contexts')


@uses_x
@browsers(FF, CH)
def test_around(tmp_path, browser):
    test_url = "about:blank"
    with _test_helper(tmp_path, index_hypothesis, test_url, browser=browser) as h:
        ts = int(datetime.strptime("2017-05-22T10:59:00.082375+00:00", '%Y-%m-%dT%H:%M:%S.%f%z').timestamp())
        h.open_page(f'search.html?timestamp={ts}')
        confirm('you should see search results, "anthrocidal" should be highlighted red')


# TODO skip if not my hostname
@uses_x
@browsers(FF, CH)
def test_chrome_visits(tmp_path, browser):
    test_url = "https://en.wikipedia.org/wiki/Amplituhedron"
    test_url = "https://en.wikipedia.org/wiki/Symplectic_vector_space"
    with _test_helper(tmp_path, index_local_chrome, test_url, browser=browser) as helper:
        trigger_command(helper.driver, Command.ACTIVATE)
        confirm("You shoud see chrome visits now; with time spent")


@uses_x
@browsers(FF, CH)
def test_show_dots(tmp_path, browser):
    visited = {
        'https://en.wikipedia.org/wiki/Special_linear_group': None,
        'http://en.wikipedia.org/wiki/Unitary_group'        : None,
        'en.wikipedia.org/wiki/Transpose'                   : None,
    }
    test_url = "https://en.wikipedia.org/wiki/Symplectic_group"
    with _test_helper(tmp_path, index_urls(visited), test_url, show_dots=True, browser=browser) as helper:
        trigger_command(helper.driver, Command.SHOW_DOTS)
        confirm("You should see dots near special linear group, Unitary group, Transpose")


@uses_x
@browsers(FF, CH)
def test_search(tmp_path, browser):
    test_url = "https://en.wikipedia.org/wiki/Symplectic_vector_space"
    with _test_helper(tmp_path, index_local_chrome, test_url, browser=browser) as helper:
        trigger_command(helper.driver, Command.SEARCH)
        # TODO actually search something?
        # TODO use current domain as deafult? or 'parent' url?
        confirm("You shoud see search prompt now, with focus on search field")


@uses_x
@browsers(FF, CH)
def test_new_background_tab(tmp_path, browser):
    start_url = "http://www.e-flux.com/journal/53/59883/the-black-stack/"
    # bg_url_text = "El Proceso (The Process)"
    # TODO generate some fake data instead?
    with _test_helper(tmp_path, index_hypothesis, start_url, browser=browser) as helper:
        confirm('you should see notification about contexts')
        helper.driver.find_element(By.XPATH, '//div[@class="logo"]/a').send_keys(Keys.CONTROL + Keys.ENTER)
        confirm('you should not see any new notifications')
        # TODO switch to new tab?
        # TODO https://www.e-flux.com/journal/53/


@uses_x
# TODO shit disappears on chrome and present on firefox
@browsers(FF, CH)
def test_local_page(tmp_path, browser):
    tutorial = 'file:///usr/share/doc/python3/html/tutorial/index.html'
    urls = {
         tutorial                                                : 'TODO read this',
        'file:///usr/share/doc/python3/html/reference/index.html': None,
    }
    url = PYTHON_DOC_URL
    with _test_helper(tmp_path, index_urls(urls), url, browser=browser) as helper:
        confirm('grey icon')
        helper.driver.get(tutorial)
        confirm('green icon. MANUALLY: ACTIVATE SIDEBAR!. It should open sidebar with one visit')
        helper.driver.back()
        # TODO it's always guaranteed to work? https://stackoverflow.com/questions/27626783/python-selenium-browser-driver-back
        confirm('grey icon, should be no sidebar')
        helper.driver.forward()
        confirm('green icon, sidebar visible')


@uses_x
@browsers(FF, CH)
def test_unreachable(tmp_path, browser):
    url = 'https://somenonexist1ngurl.com'
    urls = {
        url: 'some context',
    }
    with _test_helper(tmp_path, index_urls(urls), 'about:blank', browser=browser) as helper:
        try:
            helper.driver.get(url)
        except:
            # results in exception because it's unreachable
            pass
        confirm('green icon, no errors, desktop notification with contexts')


@uses_x
@browsers(FF, CH)
def test_stress(tmp_path, browser):
    page = tmp_path / 'dummy.html'
    page.write_text('''
<html>
<head>HI</head>
<body>test</body>
</html>
    ''')
    url = f'file://{page}'

    urls = [
        (url, f'ctx {i}' if i < 20 else None) for i in range(10000)
    ]
    with _test_helper(tmp_path, index_urls(urls), url, browser=browser) as helper:
        # TODO so, should the response be immediate?
        # TODO shouldn't do any work if we don't open the sidebar
        confirm('is performance reasonable?')

    

@uses_x
@browsers(FF, CH)
def test_fuzz(tmp_path, browser):
    urls = {
        'https://www.iana.org/domains/reserved': 'IANA',
        'iana.org/domains/reserved': 'IANA2',
    }
    with _test_helper(tmp_path, index_urls(urls), 'https://example.com', browser=browser) as helper:
        driver = helper.driver
        # import ipdb; ipdb.set_trace()
        tabs = 30
        for _ in range(tabs):
            driver.find_element_by_tag_name('a').send_keys(Keys.CONTROL + Keys.RETURN)

        sleep(5)
        for _ in range(tabs - 2):
            driver.close()
            sleep(0.1)
            driver.switch_to.window(driver.window_handles[0])

        def cb():
            for _ in range(10):
                send_key('Ctrl+Shift+t')
                sleep(0.1)
        trigger_callback(driver, cb)
        confirm("shouldn't result in 'unexpected error occured'; show only show single notification per page")


def trigger_sidebar_search(driver):
    driver.switch_to.default_content()
    driver.switch_to.frame('promnesia-sidebar')
    search_button = driver.find_element_by_xpath('//button[text()="Search"]')
    search_button.click()


@uses_x
@browsers(FF, CH)
def test_duplicate_background_pages(tmp_path, browser):
    url = PYTHON_DOC_URL
    with _test_helper(tmp_path, index_urls({}), url, browser=browser) as helper:
        driver = helper.driver

        trigger_command(driver, Command.ACTIVATE)
        # TODO separate test just for buttons from extension
        confirm('sidebar opened?')

        trigger_sidebar_search(driver)
        sleep(1)
        driver.switch_to.window(driver.window_handles[0])
        sleep(1)

        trigger_sidebar_search(driver)
        sleep(1)
        driver.switch_to.window(driver.window_handles[0])
        sleep(1)

        confirm('only two search pages should be opened')

        trigger_command(driver, Command.ACTIVATE)

        confirm('sidebar should be closed now')

        # TODO wtf? browser with search pages stays open after test... 

        # TODO getting this in chrome inspector while running this...
# VM2048 common.js:116 [background] [object Object]
# log @ VM2048 common.js:116
# notifyError @ VM2056 notifications.js:40
# Promise.catch (async)
# (anonymous) @ VM2056 notifications.js:49
# VM2056 notifications.js:17 Uncaught (in promise) TypeError: Cannot read property 'create' of undefined
#     at notify (VM2056 notifications.js:17)
#     at notifyError (VM2056 notifications.js:41)





import argparse
import os
import logging
import inspect
import sys
from typing import List, Tuple, Optional, Dict, Sequence, Iterable
from pathlib import Path
from datetime import datetime
from tempfile import TemporaryDirectory


from . import config
from . import server
from .misc import install_server
from .common import PathIsh, History, make_filter, get_logger, get_tmpdir
from .common import previsits_to_history, Source
from .dump import dump_histories


def _do_index() -> Iterable[Exception]:
    cfg = config.get()

    logger = get_logger()

    indexers = cfg.sources

    output_dir = cfg.output_dir
    if not output_dir.exists():
        logger.warning("OUTPUT_DIR '%s' didn't exist, creating", output_dir)
        output_dir.mkdir(exist_ok=True)

    filters = [make_filter(f) for f in cfg.FILTERS]
    for f in filters:
        History.add_filter(f) # meh..


    all_histories = []
    all_errors = []

    for extractor in indexers:
        ex = extractor
        if callable(ex):
            # lazy indexers
            ex = ex()
        assert isinstance(ex, Source)

        einfo = f'{ex.ff.__module__}:{ex.ff.__name__} {ex.args} {ex.kwargs}'

        hist, errors = previsits_to_history(ex, src=ex.src)
        all_errors.extend(errors)
        all_histories.append((einfo, hist))

    # TODO perhaps it's better to open connection and dump as we collect so it consumes less memory?
    dump_histories(all_histories)

    return all_errors


def do_index(config_file: Path):
    config.load_from(config_file)
    try:
        errors = _do_index()
    finally:
        config.reset()
    if len(list(errors)) > 0:
        sys.exit(1)


def demo_sources():
    def lazy(name: str):
        # helper to avoid failed imports etc, since people might be lacking necessary dependencies
        def inner(*args, **kwargs):
            from . import sources
            import importlib
            module = importlib.import_module('promnesia.sources' + '.' + name)
            return getattr(module, 'index')
        return inner

    return {
        # TODO ugh, this runs against merged db
        # 'chrome' : browser.chrome,
        # 'firefox': browser.firefox,
        'auto'    : lazy('auto'),
        # TODO org mode

        'takeout' : lazy('takeout'),
        'telegram': lazy('telegram'),
        'guess'   : lazy('guess')
    }



def do_demo(*, index_as: str, params: Sequence[str], port: Optional[str], config_file: Optional[Path]):
    logger = get_logger()
    from pprint import pprint
    with TemporaryDirectory() as tdir:
        outdir = Path(tdir)

        if config_file is not None:
            config.load_from(config_file)
        else:
            idx = demo_sources()[index_as]()
            src = Source(idx, *params)
            cfg = config.Config(
                OUTPUT_DIR=outdir,
                SOURCES=[src],
            )
            config.instance = cfg

        errors = _do_index()
        for e in errors:
            logger.error(e)

        if port is None:
            logger.warning("Port isn't specified, not serving!")
        else:
            server._run(port=port, db=config.get().output_dir / 'promnesia.sqlite', timezone=server.get_system_tz(), quiet=False)

        if sys.stdin.isatty():
            input("Press any key when ready")


def main():
    # TODO longer, literate description?
    from .common import setup_logger
    setup_logger(get_logger(), level=logging.DEBUG)

    F = argparse.ArgumentDefaultsHelpFormatter
    p = argparse.ArgumentParser(formatter_class=F)
    # p.add_argument('--hello', required=False, default='YES', help='alala')
    subp = p.add_subparsers(dest='mode', )
    ep = subp.add_parser('index', help='Create/update the link database', formatter_class=F)
    ep.add_argument('--config', type=Path, default=Path('config.py'))
    # TODO use some way to override or provide config only via cmdline?
    ep.add_argument('--intermediate', required=False)

    sp = subp.add_parser('serve', help='Serve a link database', formatter_class=F)
    server.setup_parser(sp)

    ap = subp.add_parser('demo', help='Demo mode: index and serve a directory in single command', formatter_class=F)
    # TODO use docstring or something?
    ap.add_argument('--port', type=str, help='Port to serve. If omitted, will only create the index without serving.', required=False)
    ap.add_argument('--config', type=Path, required=False, help='Config to run against. If omitted, will use empty config')
    ap.add_argument(
        '--as',
        choices=list(sorted(demo_sources().keys())),
        default='guess',
        help='Index the path as',
    )
    ap.add_argument('params', nargs='*', help='Optional extra params for the indexer')

    isp = subp.add_parser('install-server', help='Install server as a systemd service (for autostart)', formatter_class=F)
    install_server.setup_parser(isp)

    args = p.parse_args()

    # TODO is there a way to print full help? i.e. for all subparsers
    if args.mode is None:
        print('ERROR: Please specify a mode', file=sys.stderr)
        p.print_help(sys.stderr)
        sys.exit(1)

    # TODO maybe, it's better for server to compute intermediate represetnation?
    # the only downside is storage. dunno.
    # worst case -- could use database?

    with get_tmpdir() as tdir: # TODO??
        if args.mode == 'index':
             do_index(config_file=args.config)
        elif args.mode == 'serve':
            server.run(args)
        elif args.mode == 'demo':
            do_demo(index_as=getattr(args, 'as'), params=args.params, port=args.port, config_file=args.config)
        elif args.mode == 'install-server':
            install_server.install(args)
        else:
            raise AssertionError(f'unexpected mode {args.mode}')

if __name__ == '__main__':
    main()
